/*
 * kgraft_patch_bsc1173867
 *
 * Fix for CVE-2019-18680, bsc#1173867
 *
 *  Upstream commits:
 *  91573ae4aed0 ("net: rds: Fix NULL ptr use in rds_tcp_kill_sock"), stable-4.4
 *  afb4164d91c7 ("RDS: TCP: Refactor connection destruction to handle multiple
 *                 paths"), master
 *  2d746c93b6e5 ("rds: tcp: remove redundant function
 *                 rds_tcp_conn_paths_destroy()"), master
 *
 *  SLE12-SP2 and -SP3 commit:
 *  192401c8df0fe2883db7e651c22a489ea390a077
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commits:
 *  dab459d7782c3d35bd3ed23981d09b7cc93bcad5
 *  977ef3c4d40dea70fe1e163f1aaf806fd86a88c9
 *
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if !IS_MODULE(CONFIG_RDS_TCP)
#error "Live patch supports only CONFIG_RDS_TCP=m"
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1173867.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "rds_tcp"

/* klp-ccp: from net/rds/tcp.c */
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/in.h>
#include <linux/module.h>
#include <net/tcp.h>
#include <net/net_namespace.h>
#include <net/netns/generic.h>
#include <net/tcp.h>
/* klp-ccp: from net/rds/rds.h */
#include <net/sock.h>
#include <linux/scatterlist.h>
#include <linux/highmem.h>
#include <linux/mutex.h>

struct rds_connection {
	struct hlist_node	c_hash_node;
	__be32			c_laddr;
	__be32			c_faddr;
	unsigned int		c_loopback:1,
				c_outgoing:1,
				c_pad_to_32:30;
	struct rds_connection	*c_passive;

	struct rds_cong_map	*c_lcong;
	struct rds_cong_map	*c_fcong;

	struct rds_message	*c_xmit_rm;
	unsigned long		c_xmit_sg;
	unsigned int		c_xmit_hdr_off;
	unsigned int		c_xmit_data_off;
	unsigned int		c_xmit_atomic_sent;
	unsigned int		c_xmit_rdma_sent;
	unsigned int		c_xmit_data_sent;

	spinlock_t		c_lock;		/* protect msg queues */
	u64			c_next_tx_seq;
	struct list_head	c_send_queue;
	struct list_head	c_retrans;

	u64			c_next_rx_seq;

	struct rds_transport	*c_trans;
	void			*c_transport_data;

	atomic_t		c_state;
	unsigned long		c_send_gen;
	unsigned long		c_flags;
	unsigned long		c_reconnect_jiffies;
	struct delayed_work	c_send_w;
	struct delayed_work	c_recv_w;
	struct delayed_work	c_conn_w;
	struct work_struct	c_down_w;
	struct mutex		c_cm_lock;	/* protect conn state & cm */
	wait_queue_head_t	c_waitq;

	struct list_head	c_map_item;
	unsigned long		c_map_queued;

	unsigned int		c_unacked_packets;
	unsigned int		c_unacked_bytes;

	/* Protocol version */
	unsigned int		c_version;
	possible_net_t		c_net;
};

static void (*kgre_rds_conn_destroy)(struct rds_connection *conn);

/* klp-ccp: from net/rds/tcp.h */
struct rds_tcp_connection {

	struct list_head	t_tcp_node;
	struct rds_connection   *conn;
	struct socket		*t_sock;
	void			*t_orig_write_space;
	void			*t_orig_data_ready;
	void			*t_orig_state_change;

	struct rds_tcp_incoming	*t_tinc;
	size_t			t_tinc_hdr_rem;
	size_t			t_tinc_data_rem;

	/* XXX error report? */
	struct work_struct	t_conn_w;
	struct work_struct	t_send_w;
	struct work_struct	t_down_w;
	struct work_struct	t_recv_w;

	/* for info exporting only */
	struct list_head	t_list_item;
	u32			t_last_sent_nxt;
	u32			t_last_expected_una;
	u32			t_last_seen_una;
};

static void (*kgre_rds_tcp_listen_stop)(struct socket *);

/* klp-ccp: from net/rds/tcp.c */
static spinlock_t (*kgre_rds_tcp_conn_lock);
static struct list_head (*kgre_rds_tcp_conn_list);

static int (*kgre_rds_tcp_netid);

struct rds_tcp_net {
	struct socket *rds_tcp_listen_sock;
	struct work_struct rds_tcp_accept_w;
};

static void kgrp_rds_tcp_kill_sock(struct net *net)
{
	struct rds_tcp_connection *tc, *_tc;
	struct sock *sk;
	LIST_HEAD(tmp_list);
	struct rds_tcp_net *rtn = net_generic(net, (*kgre_rds_tcp_netid));

	(*kgre_rds_tcp_listen_stop)(rtn->rds_tcp_listen_sock);
	rtn->rds_tcp_listen_sock = NULL;
	flush_work(&rtn->rds_tcp_accept_w);
	spin_lock_irq(&(*kgre_rds_tcp_conn_lock));
	list_for_each_entry_safe(tc, _tc, &(*kgre_rds_tcp_conn_list), t_tcp_node) {
		struct net *c_net = read_pnet(&tc->conn->c_net);

		if (net != c_net)
			continue;
		list_move_tail(&tc->t_tcp_node, &tmp_list);
	}
	spin_unlock_irq(&(*kgre_rds_tcp_conn_lock));
	list_for_each_entry_safe(tc, _tc, &tmp_list, t_tcp_node) {
		/*
		 * Fix CVE-2019-18680
		 *  -3 lines, +5 lines
		 */
		if (tc->t_sock) {
			sk = tc->t_sock->sk;
			sk->sk_prot->disconnect(sk, 0);
			tcp_done(sk);
		}
		if (tc->conn->c_passive)
			(*kgre_rds_conn_destroy)(tc->conn->c_passive);
		(*kgre_rds_conn_destroy)(tc->conn);
	}
}

int kgrp_rds_tcp_dev_event(struct notifier_block *this,
			     unsigned long event, void *ptr)
{
	struct net_device *dev = netdev_notifier_info_to_dev(ptr);

	/* rds-tcp registers as a pernet subys, so the ->exit will only
	 * get invoked after network acitivity has quiesced. We need to
	 * clean up all sockets  to quiesce network activity, and use
	 * the unregistration of the per-net loopback device as a trigger
	 * to start that cleanup.
	 */
	if (event == NETDEV_UNREGISTER_FINAL &&
	    dev->ifindex == LOOPBACK_IFINDEX)
		kgrp_rds_tcp_kill_sock(dev_net(dev));

	return NOTIFY_DONE;
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "rds_tcp_conn_lock", (void *)&kgre_rds_tcp_conn_lock, "rds_tcp" },
	{ "rds_tcp_conn_list", (void *)&kgre_rds_tcp_conn_list, "rds_tcp" },
	{ "rds_tcp_netid", (void *)&kgre_rds_tcp_netid, "rds_tcp" },
	{ "rds_conn_destroy", (void *)&kgre_rds_conn_destroy, "rds" },
	{ "rds_tcp_listen_stop", (void *)&kgre_rds_tcp_listen_stop, "rds_tcp" }
};

static int kgr_patch_bsc1173867_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1173867_module_nb = {
	.notifier_call = kgr_patch_bsc1173867_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1173867_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1173867_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1173867_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1173867_module_nb);
}
